from copy import copy, deepcopy

class UndoManager():  # pass it a UserInterface object so Undo can tell UI
        # when to switch to another saved movie state.
        """ Manages undo/redo "stack" by storing the last 100 movie states
            in a list. Takes a UserInterface object for syntax. methods for
            push, undo and redo """
        def __init__(self, ui, appState = None):
            self.ui = ui
            self.undoIndex = 0 # will be 0 when populated with 1 state.
            self.modifications = 0 #
            self.undoList = []
            self.historySize = 100  # default, but really determined by
            self.appState = appState
            # AppState values passed to setHistorySize() below.
            self.push() # push initial state
        def push(self): # maybe should be called pushState or saveState?
            """ Take current movie, add to the end of a list of movie
                objects - ie, push current state onto the undo stack. """
            if self.modifications > 0:
                if self.appState.modified == False:
                    self.appState.modified = True
            self.modifications += 1
            if len(self.undoList) >= self.historySize:   # How far back our undo history can
                # go. Make this configurable.
                # if undo stack == full, dequeue from the bottom
                self.undoList.pop(0)
            # if undoIndex isn't indexing the last item in undoList,
            # ie we have redo states, remove all items after undoList[undoIndex]
            self.undoList = self.undoList[0:self.undoIndex]  # trim list
            # then add the new state to the end of the queue.
            self.undoList.append(deepcopy(self.ui.mov))
            # last item added == at the end of the list, so..
            self.undoIndex = len(self.undoList) # point index to last item
        def undo(self):
            if self.modifications > 1:
                self.modifications = self.modifications - 1
            if self.modifications == 2:
                self.appState.modified = False
            if self.undoIndex == 1: # nothing to undo
                self.ui.notify("Nothing to undo.")
                return False
            # if we're at the end of the list, push current state so we can
            # get back to it. A bit confusing.
            if self.undoIndex == len(self.undoList):
                self.push()
                self.undoIndex -= 1
            self.undoIndex -= 1
            self.ui.mov = self.undoList[self.undoIndex] # set UI movie state
            return True # succeeded
        def redo(self):
            if self.undoIndex < (len(self.undoList) -1): # we can redo
                self.undoIndex += 1 # go to next redo state
                self.modifications += 1
                self.ui.mov = self.undoList[self.undoIndex]
                if self.appState.modified == False:
                    self.appState.modified = True
            else:
                self.ui.notify("Nothing to redo.")
        def setHistorySize(self, historySize):
            """ Defines the max number of undo states we will save """
            self.historySize = historySize





# Takes clicks from gui_manager's Gui() objects and.. does stuff?

import pdb

class GuiHandler():
    def __init__(self, gui):
        self.gui = gui
        self.window = gui.window
        self.debugLine = 30 

    def got_click(self, clickType, y, x, extra=None):
        #self.window.addstr(self.debugLine, 0, f"Clicked: {x}, {y}, {clickType}.")
        # if a button was clicked, tell it:
        for button in self.gui.buttons:
            if x == button.realX:   # if it's on the line and \/ within width area
                if (y >= button.realY) and (y <= button.realY + button.width + 1):
                    #self.window.addstr(33 + z, 0, f"Clicky")
                    #pdb.set_trace()
                    button.on_click()




# Does things like: Takes a mouse click command. eg: left click on X, y.
# Figures out which widget got clicked (if any), tells the widget to run
# its onClick

# Eventually this should also do things like tell the GUI handler (curses
# or kivy) to initialize

from durdraw.durdraw_gui_manager_curses import GuiHandler

class Gui():
    def __init__(self, guiType=None, window=None):
        self.window = window
        self.handler = GuiHandler(self)
        self.widgets = []
        self.buttons = []

    def add_button(self, widget):
        self.buttons.append(widget)

    def del_button(self, widget):
        self.buttons.remove(widget)

    def got_click(self, clickType, x, y, extra=None):
        self.handler.got_click(clickType, x, y, extra=None)



import json

class Options():    # config, prefs, preferences, etc. Per movie. Separate from AppState options.
    """ Member variables are canvas X/Y size, Framerate, Video resolution, etc """
    def __init__(self, width=80, height=23):         # default options
        self.framerate = 8.0
        self.sizeX = width
        self.sizeY = height
        self.saveFileFormat = 7 # save file format version number
        # version 4 is pickle, version 5 is JSON, version 6 saves color and 
        # character encoding formats, Version 7 uses a new palette order for 16 color 
    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__,
            sort_keys=True, indent=4)




# Load character sets from files, initialize them 

import pdb

import pathlib
import xml.etree.ElementTree as ET

def hex_range_iterator(start_hex, end_hex):
    """ Iterator that takes a first and last hex number, and returns each one
        in succession.
        Used to generate character sets from unicode-provided ranges
        # Example usage:
        start_hex = "1FB00"
        end_hex = "1FBFF"
        for hex_value in hex_range_iterator(start_hex, end_hex):
            print(hex_value)
    """
    start_int = int(start_hex, 16)  # Convert start_hex to an integer
    end_int = int(end_hex, 16)      # Convert end_hex to an integer
    current_int = start_int
    while current_int <= end_int:
        #yield format(current_int, 'X')  # Convert the integer back to hexadecimal
        #yield hex(current_int)
        yield current_int
        current_int += 1

def parse_xml_blocks_file(xml_file_path):
    """ laod XML file, returns block_data object, containing all the block
        names and their first and last code point (hex value)
     """
    block_data = {}

    # Parse the XML file
    tree = ET.parse(xml_file_path)
    root = tree.getroot()

    # Iterate through the 'block' elements and extract the data
    for block_element in root.findall('.//block'):
        first_cp = block_element.get('first-cp')
        last_cp = block_element.get('last-cp')
        block_name = block_element.get('name')
        
        if block_name is not None and first_cp is not None and last_cp is not None:
            block_data[block_name] = {
                'first-cp': first_cp,
                'last-cp': last_cp
            }
    return block_data

# We can return a fullCharMap that looks like this: [ {'f1':\x1FB00, 
# 895         self.fullCharMap = [ \
# 896             # All of our unicode templates live here. Blank template:
# 897             #{'f1':, 'f2':, 'f3':, 'f4':, 'f5':, 'f6':, 'f7':, 'f8':, 'f9':, 'f10':},
# 898
# 899             # block characters
# 900             {'f1':9617, 'f2':9618, 'f3':9619, 'f4':9608, 'f5':9600, 'f6':9604, 'f7':9612, 'f8':9616, 'f9':     9632, 'f10':183 },    # ibm-pc looking block characters (but unicode instead of ascii)

def load_unicode_block(block_name: str):
    """ returns a fullCharMap """
    #xml_block_filename = 'unicode-groups.xml'
    xml_block_filename = pathlib.Path(__file__).parent.joinpath("charsets/unicode-groups.xml")
    block_data = parse_xml_blocks_file(xml_block_filename)
    if block_name in block_data:
        block_info = block_data[block_name]
        #print(f"Block: {block_name}")
        #print(f"First Code Point: {block_info['first-cp']}")
        #print(f"Last Code Point: {block_info['last-cp']}")
        first_cp = block_info['first-cp']
        last_cp = block_info['last-cp']
        charMap = []
        fKeyList = {'f1':'', 'f2':'', 'f3':'', 'f4':'', 'f5':'', 'f6':'', 'f7':'', 'f8':'', 'f9':'', 'f10':''}
        fKeyNum = 1
        # go through each code point, assign them to F1-F10, and add those f1-f10
        # sets to the larger full character map
        for code_point in hex_range_iterator(first_cp, last_cp):
            #fKeyList[f'f{fKeyNum}'] = chr(code_point)
            #fKeyList[f'f{fKeyNum}'] = f'\U{hex(code_point)}'
            fKeyList[f'f{fKeyNum}'] = code_point
            if fKeyNum == 10:
                charMap.append(fKeyList.copy())
                fKeyNum = 1
            else:
                fKeyNum += 1
            #print(str(fKeyList))
        return charMap
    else:
        #print(f"Block '{block_name}' not found in the XML data.")
        # this should not happen, so...
        return None

def get_unicode_blocks_list():
    xml_block_filename = pathlib.Path(__file__).parent.joinpath("charsets/unicode-groups.xml")
    block_data = parse_xml_blocks_file(xml_block_filename)
    block_list = list(block_data.keys())
    block_list = block_list[1:] # cut off Basic Latin, or 0000-007F, as 0000 is a null character
    return block_list

if __name__ == "__main__":
    #charMap = load_unicode_block("Chess Symbols")
    #charMap = load_unicode_block("Symbols for Legacy Computing")
    #charMap = load_unicode_block("Emoticons")
    #print(str(charMap))
    unicodeBlocksList = get_unicode_blocks_list()
    print(str(unicodeBlocksList))



# Part of Durdraw - https://github.com/cmang/durdraw
# main() entry point

import argparse
import configparser
import curses
import os
import sys
import time
import pathlib

from durdraw.durdraw_appstate import AppState
from durdraw.durdraw_ui_curses import UserInterface as UI_Curses
from durdraw.durdraw_options import Options
import durdraw.help

class ArgumentChecker:
    """ Place to hold any methods for verifying CLI arguments, beyond
        what argparse can do on its own. Call these methods via
        argparse.add_argument(.. type= ..) paramaters. """
    def undosize(size_s):
        size = int(size_s)  # because it comes as a string
        if size >= 1 and size <= 1000:
            return size
        else:
            raise argparse.ArgumentTypeError("Undo size must be between 1 and 1000.")

def main():
    DUR_VER = '0.25.3'
    DUR_FILE_VER = 7
    DEBUG_MODE = False # debug = makes debug_write available, sends verbose notifications
    durlogo = '''
       __                __
     _|  |__ __ _____ __|  |_____ _____ __ __ __
    / _  |  |  |   __|  _  |   __|  _  |  |  |  |\\
   /_____|_____|__|__|_____|__|___\____|________| |
   \_____________________________________________\|  v %s
    Press esc-h for help.
    ''' % DUR_VER
    argChecker = ArgumentChecker()
    parser = argparse.ArgumentParser()
    parserStartScreenMutex = parser.add_mutually_exclusive_group()
    parserFilenameMutex = parser.add_mutually_exclusive_group()
    parserColorModeMutex = parser.add_mutually_exclusive_group()
    parserFilenameMutex.add_argument("filename", nargs='?', help=".dur or ascii file to load")
    parserFilenameMutex.add_argument("-p", "--play", help="Just play .dur file or files, then exit",
                    nargs='+')
    #parserStartScreenMutex.add_argument("-q", "--quick", help="Skip startup screen",
    parserStartScreenMutex.add_argument("--startup", help="Show startup screen",
                    action="store_true")
    parserStartScreenMutex.add_argument("-w", "--wait", help="Pause at startup screen",
                    action="store_true")
    parserStartScreenMutex.add_argument("-x", "--times", help="Play X number of times (requires -p)",
                    nargs=1, type=int)
    parserColorModeMutex.add_argument("--256color", help="Try 256 color mode", action="store_true", dest='hicolor')
    parserColorModeMutex.add_argument("--16color", help="Try 16 color mode", action="store_true", dest='locolor')
    parser.add_argument("-b", "--blackbg", help="Use a black background color instead of terminal default", action="store_true")
    parser.add_argument("-W", "--width", help="Set canvas width", nargs=1, type=int)
    parser.add_argument("-H", "--height", help="Set canvas height", nargs=1, type=int)
    parser.add_argument("-m", "--max", help="Maximum canvas size for terminal (overrides -W and -H)", action="store_true")
    parser.add_argument("--nomouse", help="Disable mouse support",
                    action="store_true")
    parser.add_argument("--cursor", help="Cursor mode (block, underscore, or pipe)", nargs=1)
    parser.add_argument("--notheme", help="Disable theme support (use default theme)",
                    action="store_true")
    parser.add_argument("--theme", help="Load a custom theme file", nargs=1)
    #parser.add_argument("-A", "--ibmpc", "--cp437", help="Use Code Page 437 (IBM-PC/MS-DOS) block character encoding instead of Unicode. (Needs CP437 capable terminal and font)", action="store_true")
    parser.add_argument("--cp437", help="Display extended characters on the screen using Code Page 437 (IBM-PC/MS-DOS) encoding instead of Utf-8. (Requires CP437 capable terminal and font) (beta)", action="store_true")
    parser.add_argument("--export-ansi", action="store_true", help="Export loaded art to an .ansi file and exit")
    parser.add_argument("-u", "--undosize", help="Set the number of undo history states - default is 100. More requires more RAM, less saves RAM.", nargs=1, type=int)
    parser.add_argument("-V", "--version", help="Show version number and exit",
                    action="store_true")
    parser.add_argument("--debug", action="store_true", help=argparse.SUPPRESS)
    args = parser.parse_args()
    if args.version:
        print(DUR_VER)
        exit(0)
    if args.times and not args.play:
        print("-x option requires -p")
        exit(1)
    app = AppState()    # to store run-time preferences from CLI, environment stuff, etc.
    app.setDurVer(DUR_VER)
    app.setDurFileVer(DUR_FILE_VER)
    app.setDebug(DEBUG_MODE)
    if args.debug:
        app.setDebug(True)
    if args.undosize: 
        app.undoHistorySize = int(args.undosize[0])
    app.showStartupScreen=True


    term_size = os.get_terminal_size()
    #if args.width and args.width[0] > 80 and args.width[0] < term_size[0]:
    if args.width and args.width[0] > 1 and args.width[0] < term_size[0]:
        app.width = args.width[0]
    else:
        app.width = 80      # "sane" default screen size, 80x24..
    #if args.height and args.height[0] > 24 and args.height[0] < term_size[1]:
    if args.height and args.height[0] > 1 and args.height[0] < term_size[1]:
        app.height = args.height[0]
    else:
        app.height = 24 - 1
    if args.max:
        if term_size[0] > 80:
           app.width = term_size[0]
        if term_size[1] > 24:
            app.height = term_size[1] - 2
    if args.play:
        app.showStartupScreen=False
    elif not args.startup:  # quick startup is now the default behavior
        app.showStartupScreen=False
        app.quickStart = True
    if args.nomouse:
        app.hasMouse = False
    if args.notheme:
        app.themesEnabled = False
    if args.hicolor:
        app.colorMode = "256"
    if args.locolor:
        app.colorMode = "16"
    if args.cp437:
        app.charEncoding = 'cp437'
        #app.drawChar = '$'
        app.colorPickChar = app.CP438_BLOCK  # ibm-pc/cp437 ansi block character
        app.blockChar = app.CP438_BLOCK
        app.drawChar = app.CP438_BLOCK
    else:
        app.charEncoding = 'utf-8'
    durhelp_fullpath = pathlib.Path(__file__).parent.joinpath("help/durhelp.dur")
    if args.blackbg:
        app.blackbg = False
    else: app.blackbg = True

    # load configuration file and theme
    if app.loadConfigFile():
        if app.colorMode == "256":
            app.loadThemeFromConfig("Theme-256")
        else:
            app.loadThemeFromConfig("Theme-16")
    if args.theme:
        if app.colorMode == "256":
            app.loadThemeFile(args.theme[0], "Theme-256")
        else:
            app.loadThemeFile(args.theme[0], "Theme-16")
        app.customThemeFile = args.theme[0]

    if args.cursor:
        if args.cursor[0] in app.validScreenCursorModes:
            app.screenCursorMode = args.cursor[0]
        else:
            print("--cursor option requires one of the following: block, underscore, pipe")
            exit(1)

    # Load help file - first look for resource path, eg: python module dir
    durhelp_fullpath = ''
    durhelp_fullpath = pathlib.Path(__file__).parent.joinpath("help/durhelp.dur")
    durhelp256_fullpath = pathlib.Path(__file__).parent.joinpath("help/durhelp-256-long.dur")
    #durhelp256_fullpath = pathlib.Path(__file__).parent.joinpath("help/durhelp-256-page1.dur")
    durhelp256_page2_fullpath = pathlib.Path(__file__).parent.joinpath("help/durhelp-256-page2.dur")
    durhelp16_fullpath = pathlib.Path(__file__).parent.joinpath("help/durhelp-16-long.dur")
    #durhelp16_fullpath = pathlib.Path(__file__).parent.joinpath("help/durhelp-16-page1.dur")
    durhelp16_page2_fullpath = pathlib.Path(__file__).parent.joinpath("help/durhelp-16-page2.dur")
    app.durhelp256_fullpath = durhelp256_fullpath
    app.durhelp256_page2_fullpath = durhelp256_page2_fullpath
    app.durhelp16_fullpath = durhelp16_fullpath
    app.durhelp16_page2_fullpath = durhelp16_page2_fullpath
    #app.loadHelpFile(durhelp_fullpath)
    app.loadHelpFile(durhelp16_fullpath)
    app.loadHelpFile(durhelp16_page2_fullpath, page=2)
    if not app.hasHelpFile:
        durhelp_fullpath = 'durdraw/help/durhelp.dur'
        app.loadHelpFile(durhelp_fullpath)

    if app.showStartupScreen:
        print(durlogo)
        if app.hasHelpFile:
            print(f"Help file: Found in {durhelp_fullpath}")
        else:
            print("Help file: Not found")
        if app.ansiLove:
            print("ansilove = Found")
        else:
            print("ansilove = Not found (no PNG or GIF export support)")
        if app.PIL:
            print("PIL = Found")
        else:
            print("PIL = Not found (no GIF support)")
        if app.hasMouse:
            print("Mouse = Enabled")
        else:
            print("Mouse = Disabled")
        if app.configFileLoaded:
            print(f"Configuration file found: {app.configFileName}")
        else:
            print(f"Configuration file not found.")

        if app.themesEnabled:
            print(f"Theme: {app.themeName}")
        else:
            print(f"Theme: Default (none)")

        print("Undo history size = %d" % app.undoHistorySize)
        if app.width == 80 and app.height == 24:
            print("Canvas size: %i columns, %i lines (Default)" % (app.width, app.height))
        else:
            print("Canvas size: %i columns, %i lines" % (app.width, app.height))
        if args.wait:
            try:
                choice = input("Press Enter to Continue...")
            except KeyboardInterrupt:
                print("\nCaught interrupt, exiting.")
                exit(0)
        else:
            pass
            time.sleep(3)
    if args.play:
        app.playOnlyMode = True
        app.editorRunning = False
    #ui = curses.wrapper(UI_Curses, app)
    ui = UI_Curses(app)
    if app.hasMouse:
        ui.initMouse()
    if app.blackbg:
        ui.enableTransBackground()
    if args.filename:
        ui.loadFromFile(args.filename, 'dur')
    if args.play:
        # Just play files and exit
        app.drawBorders = False
        if args.times:
            app.playNumberOfTimes = args.times[0]
        for movie in args.play:
            ui.loadFromFile(movie, 'dur')
            ui.startPlaying()
            ui.stdscr.clear()
        ui.verySafeQuit()
    if args.export_ansi:
        # Export ansi and exit
        ui.saveAnsiFile(os.path.splitext(args.filename)[0] + ".ansi")
        ui.verySafeQuit()
    ui.refresh()
    ui.mainLoop()

if __name__ == "__main__":
    main()





# Handlers for durdraw_ui_widgets.py classes

import curses
import pdb
import time
import curses.panel

def curses_cursorOff():
    try:
        curses.curs_set(0)  # turn off cursor
    except curses.error:
        pass    # .. if terminal supports it.

def curses_cursorOn():
    try:
        curses.curs_set(1)  # turn on cursor
    except curses.error:
        pass    # .. if terminal supports it.

def curses_notify(screen, message, pause=False):
    #screen.cursorOff()
    #screen.clearStatusLine()
    screen.addstr(0, 0, message)
    screen.refresh()
    if pause:
        if screen.playing:
            screen.nodelay(0) # wait for input when calling getch
            screen.getch()
            screen.nodelay(1) # do not wait for input when calling getch
        else:
            screen.getch()
    if not pause:
        curses.napms(1500)
        curses.flushinp()
    #screen.clearStatusLine()
    #screen.cursorOn()
    screen.refresh()

def curses_addstr(window, y, x, text, attr=None): # addstr(y, x, str[, attr]) and addstr(str[, attr])
    """ Wraps curses addstr in a try;except, prevents addstr from
        crashing cureses if it fails """
    if not attr:
        try:
            window.addstr(y, x, text)
        except curses.error as e:
            curses_notify(window, f"Debug: Curses error in addstr(): {e.args[0]}")
            #self.testWindowSize()
    else:
        try:
            window.addstr(y, x, text, attr)
        except curses.error:
            pass    # silent ugly fail
            #curses_notify(window, f"Debug: Curses error in addstr(): {curses.error}")
            #self.testWindowSize()

class MenuHandler:
    """ hook into Curses to draw menu 
    """
    def __init__(self, menu, window, appState=None, statusBar=None):
        self.menu = menu
        self.window = window
        self.appState=appState
        self.parentWindow = self.menu.caller.window
        self.x = menu.x
        self.y = menu.y
        self.width = None
        self.height = None
        self.title = menu.title # show the title if one is set
        self.menuOriginLine = 0
        #self.rebuild()
        #self.panel = curses.panel.new_panel(self.curses_win)

    def rebuild(self):
        height = len(self.menu.items) + 2  # 2 for top and bottom border lines
        if self.title:
            height += 1
        # find widest item in list, go a few characters larger 
        width = len(max(self.menu.items, key = len)) + 4  + 7 # 4 for padding and side borders, more for shortcuts
        if self.menu.title:
            titleWidth = len(self.menu.title) + 4 
            if titleWidth> width:
                width = titleWidth
        self.width = width
        self.x = self.menu.x - height
        self.curses_win = curses.newwin(height, width, self.x, self.y)
        #self.curses_win.border()
        line = 1
        if self.title:
            line += 1
        textColor = curses.color_pair(self.appState.theme['mainColor']) | curses.A_BOLD
        buttonColor = curses.color_pair(self.appState.theme['clickColor'])
        shortcutColor = curses.color_pair(self.appState.theme['menuBorderColor'])
        borderColor = curses.color_pair(self.appState.theme['menuBorderColor'])
        menuTitleColor = curses.color_pair(self.appState.theme['menuTitleColor']) | curses.A_BOLD | curses.A_UNDERLINE
        maxX, maxY = self.parentWindow.getmaxyx()
        self.menuOriginLine = maxX - 2 - height
        # Draw a pretty border
        curses_addstr(self.curses_win, 0, 0, ".", borderColor)
        curses_addstr(self.curses_win, 0, 1, ("." * (self.width - 2)), borderColor)
        curses_addstr(self.curses_win, 0, width - 1, ".", borderColor)
        if self.title:
            curses_addstr(self.curses_win, 1, 0, ':', borderColor)
            curses_addstr(self.curses_win, 1, 2, self.menu.title, menuTitleColor) # Menu title
            curses_addstr(self.curses_win, 1, width - 1, ':', borderColor)
        for (item, button) in zip(self.menu.items, self.menu.buttons):
            shortcut = self.menu.items[item]["shortcut"]
            has_submenu = self.menu.items[item]["has_submenu"]

            curses_addstr(self.curses_win, line, 0, ':', borderColor)
            curses_addstr(self.curses_win, line, width - 1, ':', borderColor)
            curses_addstr(self.curses_win, line, 2, item, textColor)    # Menu item
            if shortcut:
                curses_addstr(self.curses_win, line, width - 7, shortcut, shortcutColor)
            if has_submenu:
                curses_addstr(self.curses_win, line, width - 2, ">", shortcutColor)
            top_of_menu = self.menu.caller.y - len(self.menu.buttons)
            button.update_real_xy(x=self.menuOriginLine + line, y=self.menu.y) # working for putting menu on first line
            button.window = self.window
            line += 1
        curses_addstr(self.curses_win, line, 0, ':', borderColor)
        curses_addstr(self.curses_win, line,  width - 1, ':', borderColor)
        curses_addstr(self.curses_win, line, 1, ("." * (width - 2)), borderColor)
        self.panel = curses.panel.new_panel(self.curses_win)
        try:
            self.panel.hide()
        except:
            pass

    def show(self):
        self.rebuild()
        self.panel.top()
        #self.panel.move(0,0)
        #self.panel.move(self.menuOriginLine, 0)
        #self.panel.move(self.menuOriginLine, self.menu.x)
        self.panel.move(self.menuOriginLine, self.menu.y)
        self.panel.show()
        self.curses_win.keypad(True)
        curses.panel.update_panels()
        curses.doupdate()
        self.curses_win.refresh()
        # Input loop
        self.window.nodelay(1)
        #self.curses_win.nodelay(0)
        prompting = True
        options = []
        current_option = 0
        for item in self.menu.items:
            options.append(item)
        self.refresh()
        curses_cursorOff()
        #pdb.set_trace()
        #curses.mousemask(1)
        #print('\033[?1003l') # disable mouse reporting
        borderColor = curses.color_pair(self.appState.theme['borderColor'])
        menuItemColor = curses.color_pair(self.appState.theme['menuItemColor'])
        response = "Close"  # default thing to do when done, returned to menu wrapper
        while(prompting):
            time.sleep(0.01)
            line = 1
            if self.title:
                line += 1
            c = self.window.getch()
            for item in self.menu.items:
                if item == options[current_option]: # selected item
                    textColor = menuItemColor | curses.A_REVERSE
                else:
                    textColor = menuItemColor
                curses_addstr(self.curses_win, line, 2, item, textColor)

                hotkeyIndex = 0
                itemString = next(iter(item))
                foundHotkey = False
                for letter in item:
                    if letter.lower() == self.menu.items[item]["hotkey"] and foundHotkey == False:
                        curses_addstr(self.curses_win, line, 2 + hotkeyIndex, letter, textColor | curses.A_UNDERLINE | curses.A_BOLD)
                        foundHotkey = True
                    hotkeyIndex += 1

                line += 1
                curses.panel.update_panels()
                self.window.refresh()
                if c == ord(self.menu.items[item]["hotkey"]):    # hotkey pressed
                    if self.menu.items[item]["has_submenu"]:    # If it opens a sub-menu..
                        # Keep it on the screen.
                        # Redraw previously selected as normal:
                        curses_addstr(self.curses_win, current_option + 1, 2, options[current_option], menuItemColor)
                        # Then highlight the new one.
                        current_option = options.index(item)
                        #self.rebuild()
                        textColor = menuItemColor | curses.A_REVERSE
                        curses_addstr(self.curses_win, line - 1, 2, item, textColor)
                    else:
                        # Otherwise, hide it
                        self.hide()
                    if not self.menu.caller.caller.playing:    # caller.caller is the main UI thing
                        self.window.nodelay(0)
                    self.menu.items[item]["on_click"]()
                    prompting = False
            if c == curses.KEY_UP:
                current_option = max(0, current_option - 1)
                #pdb.set_trace()
            elif c == curses.KEY_DOWN:
                current_option = min(len(options) - 1, current_option + 1)
            elif c in [13, curses.KEY_ENTER]:    # selected an option
                #pdb.set_trace()
                self.hide()
                prompting = False
                # yikes lol
                self.menu.items[options[current_option]]["on_click"]() 
                self.appState.colorPickerSelected = False
                #if not self.menu.caller.caller.playing:    # caller.caller is the main UI thing
                #    self.window.nodelay(0)
            elif c in [98, curses.KEY_LEFT]:
                self.hide()
                prompting = False
                response = "Left"
                #self.hide()
                #prompting = False
                # Here: Launch a different menu
                #self.menu.statusBar.menuButton.on_click
            elif c in [102, curses.KEY_RIGHT]:
                if self.menu.items[options[current_option]]["has_submenu"]:    # If it opens a sub-menu..
                    #curses_notify(window, f"Debug: Fnord")
                    #self.hide()
                    #prompting = False
                    self.menu.items[options[current_option]]["on_click"]() 
                    prompting = False
                else:
                    # Jump out of the menu, and tell the parent handler to move to the next menu
                    self.hide()
                    prompting = False
                    response = "Right"
                    #self.hide()
                    #prompting = False
                    # Here: Launch a different menu
            elif c == 27:  # normal esc
                self.hide()
                prompting = False
            elif c == curses.KEY_MOUSE:
                try:
                    _, mouseX, mouseY, _, mouseState = curses.getmouse()
                except:
                    pass

                if not self.appState.hasMouseScroll:
                    curses.BUTTON5_PRESSED = 0
                    curses.BUTTON4_PRESSED = 0 
                if mouseState & curses.BUTTON4_PRESSED:   # wheel up
                    current_option = max(0, current_option - 1)
                elif mouseState & curses.BUTTON5_PRESSED:   # wheel down
                    current_option = min(len(options) - 1, current_option + 1)
                else:   # assume a click
                    # Did the user click in the menu area?
                    if mouseY > self.menuOriginLine and mouseY < self.menuOriginLine + len(self.menu.items):  # on a menu line?
                        if mouseX < self.x and mouseX > self.x - self.width:    # in a menu column
                            # Un-highlight selected item
                            curses_addstr(self.curses_win, current_option + 1, 2, options[current_option], menuItemColor)
                            # Highlight the one we're clicking on
                            current_option = mouseY - self.menuOriginLine - 1
                            #item = self.menu.items[options[current_option]]
                            #curses_notify(self.window, f"current_option: {current_option}")
                            #self.rebuild()
                            textColor = menuItemColor | curses.A_REVERSE
                            curses_addstr(self.curses_win, current_option + 1, 2, options[current_option], textColor)

                            has_submenu = self.menu.items[options[current_option]]["has_submenu"]
                            if has_submenu:
                                prompting = False
                                self.menu.items[options[current_option]]["on_click"]() 
                            else:
                                self.hide()
                                prompting = False
                                #self.menu.items[options[current_option]]["on_click"]() 
                                if not self.menu.caller.caller.playing:    # caller.caller is the main UI thing
                                    self.window.nodelay(0)
                                self.menu.gui.got_click("Click", mouseX, mouseY)
                        else:
                            #curses_notify(self.window, f"Debug: mouseX: {mouseX}, mouseY: {mouseY}, self.x: {self.x}, self.menuOriginLine: {self.menuOriginLine}")
                            prompting = False
                            self.menu.gui.got_click("Click", mouseX, mouseY)
                            self.hide()
                            #curses_notify(self.window, f"This should never happen. mouseX: {mouseX}, mouseY: {mouseY}, self.x: {self.x}, self.menuOriginLine: {self.menuOriginLine}")
                                
                    else:
                        #curses_notify(self.window, f"Debug: mouseX: {mouseX}, mouseY: {mouseY}, self.x: {self.x}, self.menuOriginLine: {self.menuOriginLine}")
                        prompting = False
                        self.menu.gui.got_click("Click", mouseX, mouseY)
                        self.hide()
            #jif c in [104, 63]:  # h H Help
            #    self.hide()
            #    self.items["Help"]["on_click"]()
            #    prompting = False
        #pdb.set_trace()
        curses_cursorOn()
        self.menu.hide()
        if not self.menu.caller.caller.playing:    # lol .. the caller.caller is the main UI thing
            self.window.nodelay(0)
        #pdb.set_trace()
        return response
        #curses_addstr(self.window, self.menu.x, self.menu.y, "Show menu")

    def refresh(self):
        self.window.refresh()
        curses.panel.update_panels()
        #self.window.move(0,0)
        curses.doupdate()

    def hide(self):
        try:
            self.panel.hide()
        except:
            pass 
        self.refresh()
        #curses_addstr(self.window, self.menu.x, self.menu.y, "Hide menu")

class DrawCharPickerHandler:
    def __init__(self, caller, window):
        self.caller = caller    # drawCharPicker
        self.window = window

    def pickChar(self):
        maxLines, maxCol = self.window.getmaxyx()
        #pdb.set_trace()
        self.window.addstr(maxLines - 3, 0, "Enter a character to use for drawing: ")
        prompting = True
        while prompting:
            c = self.window.getch()
            time.sleep(0.01)
            if c in [curses.KEY_F1]:
                self.caller.appState.drawChar = chr(self.caller.caller.caller.chMap['f1'])
                prompting = False
            elif c in [curses.KEY_F2]:
                self.caller.appState.drawChar = chr(self.caller.caller.caller.chMap['f2'])
                prompting = False
            elif c in [curses.KEY_F3]:
                self.caller.appState.drawChar = chr(self.caller.caller.caller.chMap['f3'])
                prompting = False
            elif c in [curses.KEY_F4]:
                self.caller.appState.drawChar = chr(self.caller.caller.caller.chMap['f4'])
                prompting = False
            elif c in [curses.KEY_F5]:
                self.caller.appState.drawChar = chr(self.caller.caller.caller.chMap['f5'])
                prompting = False
            elif c in [curses.KEY_F6]:
                self.caller.appState.drawChar = chr(self.caller.caller.caller.chMap['f6'])
                prompting = False
            elif c in [curses.KEY_F7]:
                self.caller.appState.drawChar = chr(self.caller.caller.caller.chMap['f7'])
                prompting = False
            elif c in [curses.KEY_F8]:
                self.caller.appState.drawChar = chr(self.caller.caller.caller.chMap['f8'])
                prompting = False
            elif c in [curses.KEY_F9]:
                self.caller.appState.drawChar = chr(self.caller.caller.caller.chMap['f9'])
                prompting = False
            elif c in [curses.KEY_F10]:
                self.caller.appState.drawChar = chr(self.caller.caller.caller.chMap['f10'])
                prompting = False
            elif c == 27:   # esc, cancel
                prompting = False
            else:
                self.caller.appState.drawChar = chr(c)
                prompting = False
        #self.caller.caller.drawCharPickerButton.label = self.caller.appState.drawChar
        self.caller.caller.drawCharPickerButton.set_label(self.caller.appState.drawChar)
        self.window.addstr(maxLines - 3, 0, "                                          ")
        self.caller.caller.caller.refresh()

class ColorPickerHandler:
    def __init__(self, colorPicker, window):
        self.colorPicker = colorPicker
        self.parentWindow = window
        self.x = colorPicker.x
        self.y = colorPicker.y
        self.parentWindow = colorPicker.caller.stdscr
        self.appState = colorPicker.caller.appState
        # figure out picker size
        #total = curses.COLORS
        #total = curses.COLORS
        realmaxY,realmaxX = self.parentWindow.getmaxyx()
        self.realmaxY = realmaxY
        self.realmaxX = realmaxX
        #self.height = int(total / realmaxX) + 2 # enough lines to fill content+
        #self.height = 4
        self.height = 8
        #height = int(total / realmaxY) # enough lines to fill content+
        #self.width = realmaxX - 10 
        #self.width = 78
        #self.width = 38 
        self.width = 38 
        #gridLine = [[]] * self.width
        #self.colorGrid = [gridLine] * self.height
        #self.colorGrid = [[0] * self.width] * self.height
        self.colorGrid = [[0 for i in range(self.width)] for j in range(self.height)]
        self.window = curses.newwin(self.height, self.width, self.x, self.y)
        self.window.keypad(True)
        self.curses_win = self.window
        self.panel = curses.panel.new_panel(self.curses_win)
        self.panel.hide()
        #self.fillChar = 9608    # unicode block
        self.fillChar = self.appState.colorPickChar # unicode block
        self.origin = self.x - 2
        #self.move(0,self.x - 2)
        self.move(0,self.origin)

    def drawBorder(self):
        """ Draw a highlighted border around color picker to show it is selected. """
        x = self.x - 2
        y = self.y - 1
        width = self.width + 1
        borderColor = curses.color_pair(self.appState.theme['menuBorderColor'])
        curses_addstr(self.parentWindow, y, x, ("." * (width)), borderColor | curses.A_BOLD)
        for line in range(1, self.height + 1):
            curses_addstr(self.parentWindow, y + line, x, (":"), borderColor | curses.A_BOLD)

    def hideBorder(self):
        x = self.x - 2
        y = self.y - 1
        width = self.width + 1
        borderColor = curses.color_pair(self.appState.theme['menuBorderColor'])
        curses_addstr(self.parentWindow, y, x, (" " * (width)))
        for line in range(1, self.height + 1):
            curses_addstr(self.parentWindow, y + line, x, (" "))

    def show(self):
        #self.showFgPicker()
        self.updateFgPicker()
        #self.updateBgPicker()
        #prompting = False
        #print('\033[?1003l') # disable mouse movement tracking (xterm api)
        #curses.mousemask(1)
        #curses_cursorOff()
        # populate window with colors
        self.panel.top()
        #self.move(0,self.x - 6)
        self.panel.show()
        #oldColor = self.colorPicker.caller.colorfg
        #color = self.colorPicker.caller.colorfg
        #if self.appState.colorPickerSelected:
        #    prompting = True
        #else:
        #    prompting = False
        #if self.appState.colorPickerSelected:
        #    if self.appState.sideBarShowing:
        #        self.drawBorder()

    def hide(self):
        self.panel.bottom()
        try:
            self.panel.hide()
        except:
            pass

    def move(self, x, y):
        self.x = x
        self.y = y
        try:
            self.panel.move(y, x)
        except Exception as E:
            #self.colorPicker.caller.notify(f"Exception {E}")
            #pdb.set_trace()
            pass
        self.origin = self.y

    def updateFgPicker(self):
        line = 0
        col = 1
        #maxWidth = self.realmaxX
        #maxHeight = self.realmaxY
        #for fg in range(0,curses.COLORS):  # 0-255
        width_counter = 0   # for color block width
        for fg in range(1,self.appState.totalFgColors+1):  # 0-255
            color_pair = curses.color_pair(fg)
            if col >= self.width - 2:
                col = 0
                line += 1
            if fg == 16:    # first color for fancy displayed block color palette thing
                line += 1
                col = 0
            if fg > 16:
                width_counter += 1
            if width_counter == 36:
                width_counter = 0
                #line += 1
            try:
                self.colorGrid[line][col] = fg
            except IndexError:
                pass
            #if line > 0:
            #    pdb.set_trace()
            #curses_addstr(self.window, self.colorPicker.y + line, self.colorPicker.x + col, chr(self.fillChar), color_pair)
            # if fg == app's current fg, draw it as a * instead of self.fillChar
            if fg == self.colorPicker.caller.colorfg:
                if fg == 1: # black
                    plain_color_pair = curses.color_pair(9)
                    if self.appState.colorPickerSelected:
                        curses_addstr(self.window, line, col, 'X', plain_color_pair | curses.A_UNDERLINE | curses.A_BLINK)
                    else:
                        curses_addstr(self.window, line, col, 'X', plain_color_pair)
                else:
                    if self.appState.colorPickerSelected:
                        curses_addstr(self.window, line, col, 'X', color_pair | curses.A_UNDERLINE | curses.A_BLINK)
                    else:
                        curses_addstr(self.window, line, col, 'X', color_pair)
            else:
                curses_addstr(self.window, line, col, self.fillChar, color_pair)
            col += 1
        curses_addstr(self.window, line, col + 5, "     ", color_pair)
        curses_addstr(self.window, line, col + 5, str(self.colorPicker.caller.colorfg), color_pair)

    def showFgPicker(self):
        #self.colorPicker.caller.notify(f"showFgPicker")
        self.showColorPicker(type="fg")

    def showColorPicker(self, type="fg"):
        #self.colorPicker.caller.notify(f"showColorPicker")
        """ Shows picker, has UI loop for letting user pick color with keyboard or mouse """
        if type == "fg":
            self.updateFgPicker()
        elif type == "bg":
            self.updateBgPicker()
        prompting = False
        #print('\033[?1003l') # disable mouse movement tracking (xterm api)
        #curses.mousemask(1)
        curses_cursorOff()
        # populate window with colors
        self.panel.top()
        #self.move(0,self.x - 6)
        self.panel.show()
        oldColor = self.colorPicker.caller.colorfg
        color = self.colorPicker.caller.colorfg
        if self.appState.colorPickerSelected:
            prompting = True
        else:
            prompting = False
        if self.appState.colorPickerSelected:
            if self.appState.sideBarShowing:
                self.drawBorder()
        #self.window.nodelay(1)

        #self.colorPicker.caller.notify(f"showColorPicker() hit. {prompting=}")
        
        while(prompting):
            time.sleep(0.01)
            #self.colorPicker.caller.drawStatusBar()
            self.update()
            c = self.window.getch()
            if c in [98, curses.KEY_LEFT]:
                if color == 0:
                    color = curses.COLORS - 1
                else:
                    color -= 1
                #self.colorPicker.caller.colorfg = color
                self.colorPicker.caller.setFgColor(color)
                self.updateFgPicker()
                self.colorPicker.caller.drawStatusBar()
            elif c in [102, curses.KEY_RIGHT]:
                color += 1
                if color >= curses.COLORS:
                    color = 0
                #self.colorPicker.caller.colorfg = color
                self.colorPicker.caller.setFgColor(color)
                self.updateFgPicker()
                self.colorPicker.caller.drawStatusBar()
            elif c == curses.KEY_UP:
                if color < 32:
                    color -= 16
                elif color > 31 and color < 52:
                    color = 15
                else:
                    color -= self.width - 2
                if color <= 0:
                    color = 1
                self.colorPicker.caller.setFgColor(color)
                self.updateFgPicker()
                self.colorPicker.caller.drawStatusBar()
            elif c == curses.KEY_DOWN:
                if color < 16:
                    color += 16
                else:
                    color += self.width - 2
                if color >= curses.COLORS:
                    color = curses.COLORS - 1
                self.colorPicker.caller.setFgColor(color)
                self.updateFgPicker()
                self.colorPicker.caller.drawStatusBar()
            elif c == curses.KEY_HOME:
                color = 1
                self.colorPicker.caller.setFgColor(color)
                self.updateFgPicker()
                self.colorPicker.caller.drawStatusBar()
            elif c == curses.KEY_END:
                color = 255
                self.colorPicker.caller.setFgColor(color)
                self.updateFgPicker()
                self.colorPicker.caller.drawStatusBar()
            elif c in [13, curses.KEY_ENTER, 9, 353]:   # Return, Accept color. 9=tab, 353=shift-tab
                #if not self.appState.stickyColorPicker:
                if not self.appState.sideBarShowing:
                    self.hide()
                prompting = False
                self.appState.colorPickerSelected = False
                c = None
                self.updateFgPicker()
                #self.colorPicker.caller.notify(f"{c=}, {prompting=}")
                if not self.colorPicker.caller.playing:    # caller.caller is the main UI thing
                    self.window.nodelay(0)  # wait
                #return color
            elif c == curses.KEY_MOUSE:
                try:
                    _, mouseX, mouseY, _, mouseState = curses.getmouse()
                except:
                    pass

                if not self.appState.hasMouseScroll:
                    curses.BUTTON5_PRESSED = 0
                    curses.BUTTON4_PRESSED = 0
                if mouseState & curses.BUTTON4_PRESSED: # wheel up
                    color += 1
                    if color >= curses.COLORS:
                        color = 0
                    #self.colorPicker.caller.colorfg = color
                    self.colorPicker.caller.setFgColor(color)
                    self.updateFgPicker()
                    self.colorPicker.caller.drawStatusBar()
                elif mouseState & curses.BUTTON5_PRESSED:   # wheel down
                    if color == 0:
                        color = curses.COLORS - 1
                    else:
                        color -= 1
                    #self.colorPicker.caller.colorfg = color
                    self.colorPicker.caller.setFgColor(color)
                    self.updateFgPicker()
                    self.colorPicker.caller.drawStatusBar()
                elif mouseY >= self.origin and mouseX > self.x and mouseX < self.x + len(self.colorGrid[0])-2:   # cpicked in the color picker
                    #self.colorPicker.caller.notify(f"DEBUG: self.origin={self.origin}, self.x = {self.x}. mouseX={mouseX}, mouseY={mouseY}", pause=True)
                    self.gotClick(mouseX, mouseY)
                    prompting = False

                    #clickedCol = mouseX - self.x
                    #clickedLine = mouseY - self.origin
                    #if mouseY < self.origin + self.height:
                    #    if self.colorGrid[clickedLine][clickedCol] != 0:
                    #        # We're in the grid. Set color
                    #        color = self.colorGrid[clickedLine][clickedCol]
                    #        self.colorPicker.caller.setFgColor(color)
                    #        self.updateFgPicker()



                if not prompting:
                    if not self.appState.sideBarShowing:
                        self.hide()
                #self.hide()
                #prompting = False
            elif c == 27:  # normal esc, Cancel
                c = self.window.getch()
                if c == curses.ERR: # Just esc was hit, no other escape sequence
                    self.colorPicker.caller.setFgColor(oldColor)
                    self.updateFgPicker()
                    if not self.appState.sideBarShowing:
                        self.hide()
                    #self.hide()
                    prompting = False

        if not self.appState.colorPickerSelected:
            if self.appState.sideBarShowing:
                self.hideBorder()

        self.appState.colorPickerSelected = False   # done prompting
        self.updateFgPicker()

        if not self.appState.sideBarShowing:
            self.hide()

        #self.hide()
        curses_cursorOn()
        self.window.nodelay(0)
        #curses.mousemask(curses.REPORT_MOUSE_POSITION | curses.ALL_MOUSE_EVENTS)
        #print('\033[?1003h') # enable mouse tracking
        return color

    def gotClick(self, mouseX, mouseY):
        if mouseY >= self.origin and mouseX < + self.x + len(self.colorGrid[0])-2:   # cpicked in the color picker
            clickedCol = mouseX - self.x
            clickedLine = mouseY - self.origin
            if mouseY < self.origin + self.height:
                if self.colorGrid[clickedLine][clickedCol] != 0:
                    color = self.colorGrid[clickedLine][clickedCol]
                    self.colorPicker.caller.setFgColor(color)
                    self.updateFgPicker()


    def update(self):
        curses.panel.update_panels()
        curses.doupdate()
        self.window.refresh()

class ColorSwatchHandler:
    def __init__(self, colorSwatch, window):
        self.colorSwatch = colorSwatch
        self.window = window
        self.fillChar = 9608    # unicode block
    def draw(self, plusX=0, plusY=0):
        for c in range(0, len(self.colorSwatch.bank)-1):
            curses_addstr(self.window, self.colorSwatch.y, self.colorSwatch.x, chr(self.fillChar))

class FgBgColorPickerHandler:
    def __init__(self, fgBgPicker, x=0, y=0):
        self.fgBgPicker = fgBgPicker
        self.window = fgBgPicker.window 
        self.x = x
        self.y = y
        #self.on_click = on_click
    def draw(self, plusX=0, plusY=0):
        curses_addstr(self.window, self.y, self.x, "F:  G:  ")

class ToolTipHandler:
    """ Draw and hide tooltips """
    #def __init__(self, tooltip, window, appState=None):
    def __init__(self, tooltip, context):
        self.tooltip = tooltip
        self.window = context.window
        self.appState = context.appState

    def draw(self):
        tipString = self.tooltip.hotkey
        tipColor = self.appState.theme['clickHighlightColor'] | curses.A_BOLD | curses.A_UNDERLINE
        #curses_addstr(self.window, self.tooltip.column, self.tooltip.row, tipString, tipColor)
        curses_addstr(self.window, self.tooltip.row, self.tooltip.column, tipString, tipColor)

    def show(self):
        self.draw()

    def hide (self):
        pass


class ButtonHandler:
    """ hook into Curses to draw button
    """
    def __init__(self, button, window, on_click, appState=None):
        #self.label = button.label
        self.button = button
        self.window = window
        self.x = self.button.x
        self.y = self.button.y
        self.on_click = on_click
        self.appState = appState
        self.color = curses.color_pair(self.appState.theme['clickColor']) | curses.A_BOLD   # bright green, clickable

    def draw(self, plusX=0, plusY=9):
        if not self.button.invisible:
            textColor = self.color
            if self.button.selected:
                #curses_notify(self.window, "Tacos")
                textColor = textColor | curses.A_REVERSE
            if self.button:   # a selector type button, to be iframed in []s
                buttonString = f"[{self.button.label}]"
            else:
                buttonString = self.button.label
            #curses_addstr(self.window, plusX + self.button.x, plusY + self.button.y, self.button.label)
            #curses_addstr(self.window, plusX + self.button.x, plusY + self.button.y, self.button.label, self.color)
            #curses_addstr(self.window, self.button.realX, self.button.realY, self.button.label, textColor)
            curses_addstr(self.window, self.button.realX, self.button.realY, buttonString, textColor)
            # render the button on the window
            if self.button.persistant_tooltip or not self.button.tooltip_hidden:
                if self.button.get_tooltip_command():
                    toolTip :str = self.button.get_tooltip_command()
                    tipColor = self.appState.theme['clickHighlightColor'] | curses.A_BOLD | curses.A_UNDERLINE
                    # Figure out if the hint is in the button label, if so, place it over it with
                    # an index offset
                    tipColOffset = 0
                    if toolTip.lower() in self.button.label.lower():
                        tipColOffset = self.button.label.lower().index(toolTip.lower()) + 1
                        #curses_notify(self.window, "Gorditas")
                    # keep the tip from going off screen for some buttons
                    if self.button.realY == 0 and tipColOffset == 0:
                        tipColOffset += 1
                    # Print it next to the button for now
                    curses_addstr(self.window, self.button.realX, self.button.realY + tipColOffset, toolTip, tipColor)


    def showToolTip(self):
        self.button.tooltip_hidden = False

    def hideToolTip(self):
        self.button.tooltip_hidden = True
        # Cover up with spaces
        #if not self.button.hidden:
        #    self.button.draw()
        #if self.button.get_tooltip_command() and not self.button.hidden:
        #    toolTip :str = self.button.get_tooltip_command()
        #    toolTip = " " * len(toolTip)
        #    curses_addstr(self.window, self.button.realX, self.button.realY, toolTip)

    def hide(self):
        self.hidden = True

    def on_click(self):
        curses_addstr(self.window, 0, 0, f"{self.button.label} clicked")

    def update_real_xy(x=0, y=0):
        self.button.realX = x
        self.button.realY = y

    def handle_event(self, event):
        # handle events for the button
        if event.type == 'click':
            self.on_click()
        return True

class StatusBarHandler:
    def __init__(self, statusBar, window): 
        self.statusBar = statusBar
        self.window = window
        self.panel = curses.panel.new_panel(self.window)
        #curses_addstr(window, self.statusBar.x, self.statusBar.y, "Fnord")
        #self.on_click = on_click

    def draw(self, plusX=0, plusY=0):
        pass
        #for item in self.statusBar.items:
        #    curses_addstr(self.window, self.statusBar.x + item.x, self.statusBar.y + item.y, item.label)

    def handle_event(self, event):
        # handle events for the button
        if event.type == 'click':
            self.on_click()
        return True






#!/usr/bin/env python3 

# Theory of operation notes:
# A code starts with '\x1B[' or ^[ and ends with a LETTER.
# The ltter is usually 'm' (for Select Graphic Rendition).
# Note that it may end with a letter other than 'm' for some features,
# like cursor movements. For example:
# ^[3A means "move cursor up 3 lines"
# A code looks like this: ^[1;32m for bold color 32. ^[0m for reset.
# ^[0;4;3;42m for reset (0), underline (4), italic (3), background color (42).
# We have a state machine that has attributes. Color, bold, underline, italic, etc.
# 
# References: 
# Escape code bible: https://en.wikipedia.org/wiki/ANSI_escape_code
# ANSI Sauce (metadata) spec: https://www.acid.org/info/sauce/sauce.htm


import sys
import struct
import re
import pdb
import durdraw.durdraw_movie as durmovie
import durdraw.durdraw_color_curses as dur_ansilib
import durdraw.durdraw_sauce as dursauce

def ansi_color_to_durcolor(ansiColor):
    colorName = ansi_color_to_durcolor_table[ansiColor]
    durColor = color_name_to_durcolor_table[colorName]
    return durColor

ansi_color_to_durcolor_table = {
    # foreground
    '30': 'Black',
    '31': 'Red',
    '32': 'Green',
    '33': 'Yellow',
    '34': 'Blue',
    '35': 'Magenta',
    '36': 'Cyan',
    '37': 'White',
    # background
    '40': 'Black',
    '41': 'Red',
    '42': 'Green',
    '43': 'Yellow',
    '44': 'Blue',
    '45': 'Magenta',
    '46': 'Cyan',
    '47': 'White'
}

color_name_to_durcolor_table = {
    'Black': 0,
    'Red': 5,
    'Green': 3,
    'Yellow': 7,
    'Blue': 2,
    'Magenta': 6,
    'Cyan': 4,
    'White': 8,
    
    'Black': 00,
    'Red': 00,
    'Green': 00,
    'Yellow': 00,
    'Blue': 00,
    'Magenta': 00,
    'Cyan': 00,
    'White': 00
}


def get_width_and_height_of_ansi_blob(text, width=80):
    i = 0   # index into the file blob
    col_num = 0
    line_num = 0
    max_col = 0
    while i < len(text):
        # If there's an escape code, extract data from it
        if text[i:i + 2] == '\x1B[':    # Match ^[
            match = re.search('[a-zA-Z]', text[i:]) # match any control code 
            end_index = match.start() + i   # where the code ends
            if text[end_index] == 'A':      # Move the cursor up X spaces
                escape_sequence = text[i + 2:end_index]
                if len(escape_sequence) == 0:
                    escape_sequence = 1
                move_by_amount = int(escape_sequence)
                line_num = line_num - move_by_amount
                i = end_index + 1
                continue    # jump the while
            elif text[end_index] == 'B':      # Move the cursor down X spaces
                escape_sequence = text[i + 2:end_index]
                if len(escape_sequence) == 0:
                    escape_sequence = 1
                move_by_amount = int(escape_sequence)
                line_num += move_by_amount
                i = end_index + 1
                continue    # jump the while
            elif text[end_index] == 'C':      # Move the cursor forward X spaces
                escape_sequence = text[i + 2:end_index]
                if len(escape_sequence) == 0:
                    escape_sequence = 1
                move_by_amount = int(escape_sequence)
                col_num += move_by_amount
                i = end_index + 1
                continue    # jump the while
            elif text[end_index] == 'D':      # Move the cursor back X spaces
                escape_sequence = text[i + 2:end_index]
                if len(escape_sequence) == 0:
                    escape_sequence = 1
                move_by_amount = int(escape_sequence)
                col_num = col_num - move_by_amount
                i = end_index + 1
                continue    # jump the while
            elif text[end_index] == 'H':      # Move the cursor to row/column
                escape_sequence = text[i + 2:end_index]
                escape_codes = escape_sequence.split(';')
                if len(escape_codes) > 1:   # row ; column
                    if escape_codes[0].isnumeric():
                        line_num = int(escape_codes[0])
                    if escape_codes[1].isnumeric():
                        col_num = int(escape_codes[1])
                elif len(escape_codes) == 1:   # row, column=1
                    #line_num = 1
                    if escape_codes[0].isnumeric():
                        col_num = int(escape_codes[0])
                i = end_index + 1
                continue    # jump the while
            elif text[end_index] == 'J':      # Clear screen
                # 0 or none = clear from cursor to end of screen
                # 1 = from cursor to top of screen
                # 2 = clear screen and move cursor to top left
                # 3 = clear entire screen and delete all lines saved in the scrollback buffer
                escape_sequence = text[i + 2:end_index]
                if len(escape_sequence) == 0:
                    escape_sequence = '0'   # default - clear from cursor to end
                if escape_sequence == '2':  # cls, move to top left
                    # using a sledgehammer to claer the screen
                    #new_frame = durmovie.Frame(width, height + 1)
                    col_num, line_num = 0, 0
                i = end_index + 1   # move on to next byte
                continue
            elif text[end_index] == 's':    # save current position/state
                saved_col_num = col_num
                saved_line_num = line_num
            elif text[end_index] == 'u':    # restore saved position/state
                col_num = saved_col_num
                line_num = saved_line_num
            i = end_index + 1
            continue    # jump the while
        # Or, not an escape character
        elif text[i] == '\n':   # new line (LF)
            line_num += 1
            if col_num > max_col:
                max_col = col_num
            col_num = 0
        elif text[i] == '\r':  # windows style newline (CR)
            pass    # pfft
        elif text[i:i + 5] == 'SAUCE' and len(text) - i == 128:   # SAUCE record found
            i += 128
            # Wee, I'm flying
        else:   # printable character (hopefully)
            if col_num == width:
                col_num = 0
                line_num += 1
            character = text[i]
            character = text[i]
            col_num += 1
        i += 1
    print("")
    width = max_col
    height = line_num
    return width, height

def parse_ansi_escape_codes(text, filename = None, appState=None, caller=None, console=False, debug=False, maxWidth=80):
    """ Take an ANSI file blob, load it into a DUR frame object, return 
        frame """
    if filename:
        # If we can just pull it from the Sauce, cool
        sauce = dursauce.SauceParser()
        sauce.parse_file(filename)
        if sauce.sauce_found:
            appState.sauce = sauce
            #if sauce.height > 0 and sauce.width > 0:
            #if sauce.height == None:
            #    sauce.height = 25
            if sauce.width == None:
                sauce.width = 80
            maxWidth = sauce.width
            width = sauce.width
            height = sauce.height
            #caller.notify(f"Sauce pulled: author: {sauce.author}, title: {sauce.title}, width {width}, height {height}")
    if not sauce.height:
        width, height = get_width_and_height_of_ansi_blob(text, width=80)
        width = sauce.width
    if not sauce.sauce_found or width > 200 or height > 1200:   # let the dodgy function guess
        width, height = get_width_and_height_of_ansi_blob(text, width=80)
    #width = max(width, maxWidth)
    width = max(width, 80)
    height += 1
    if appState.debug:
        caller.notify(f"Guessed width: {width}, height: {height}")
    #width = min(width, maxWidth)
    height = max(height, 25)
    new_frame = durmovie.Frame(width, height + 1)
    if appState.debug:
        caller.notify(f"debug: maxWidth = {maxWidth}")
    #parsed_text = ''
    #color_codes = ''
    i = 0   # index into the file blob
    col_num = 0
    line_num = 0
    max_col = 0
    default_fg_color = appState.defaultFgColor
    default_bg_color = appState.defaultBgColor
    fg_color = default_fg_color 
    bg_color = default_bg_color 
    bold = False
    saved_col_num = 0
    saved_line_num = 0
    saved_byte_location = 0
    parse_error = False
    while i < len(text):
        # If there's an escape code, extract data from it
        if text[i:i + 2] == '\x1B[':    # Match ^[[
            match = re.search('[a-zA-Z]', text[i:]) # match any control code 
            end_index = match.start() + i   # where the code ends
            if text[end_index] == 'm':      # Color/SGR control code
                escape_sequence = text[i + 2:end_index]
                escape_codes = escape_sequence.split(';')
                codeList = []
                for code in escape_codes:
                    try:
                        codeList.append(int(code))
                    except:
                        if caller:
                            pass
                            #caller.notify(f"Error in byte {i}, char: {code}, line: {line_num}, col: {col_num}")
                if len(codeList) > 1 and appState.colorMode == "256":
                    # 256 foreground color
                    if codeList[0] == 38 and codeList[1] == 5 and len(codeList) == 3:
                        fg_color = codeList.pop()
                        codeList = [fg_color]
                    # 256 background color
                    elif codeList[0] == 48 and codeList[1] == 5 and len(codeList) == 3:
                        bg_color = codeList.pop()
                        codeList = [fg_color]
                # Not a 256 color code - treat as 16 color
                for code in codeList:
                    if code == 0:   # reset
                        fg_color = default_fg_color
                        bg_color = default_bg_color
                        bold = False
                    if code == 1:   # bold
                        bold = True
                    # In case we're still using a previous color with new attributes
                    if fg_color < 9:  # sledgehammer
                        if bold:
                            fg_color += 8
                    # 16 Colors
                    if code > 29 and code < 38: # FG colors 0-8, or 30-37
                        if bold:
                            code += 60  # 30 -> 90, etc, for DOS-style bright colors that use bold
                            #bold = False
                        if appState.colorMode == "256":
                            fg_color = dur_ansilib.ansi_code_to_dur_16_color[str(code)] - 1
                        else:
                            #if bold:
                            #    code += 60  # 30 -> 90, etc, for DOS-style bright colors that use bold
                            fg_color = dur_ansilib.ansi_code_to_dur_16_color[str(code)] 
                            #if bold:
                            #    fg_color += 8
                        # fix for durdraw color pair stupidity
                        if fg_color == -1 or fg_color == 0: # black fg and bright black fg fix
                            if bold:
                                fg_color = 9
                            else:
                                fg_color = 1
                        #if fg_color == 8:   # bright white fix
                        #    if bold:
                        #        fg_color = 16
                        #bold = False
                        if fg_color < 9:  # sledgehammer
                            if bold:
                                fg_color += 8
                    elif code > 39 and code < 48: # BG colors 0-8, or 40-47
                        if appState.colorMode == "256":
                            #bg_color = dur_ansilib.ansi_code_to_dur_16_color[str(code)] - 1
                            #bg_color = 0
                            bg_color = dur_ansilib.ansi_code_to_dur_16_color[str(code)] - 1
                            if bg_color == -1:
                                bg_color = 0
                        else:
                            #if bold:
                            #    code += 60  # 30 -> 90, etc, for DOS-style bright colors that use bold
                            bg_color = dur_ansilib.ansi_code_to_dur_16_color[str(code)] - 1
                            if bg_color == -1:
                                bg_color = 0
                    if fg_color == -1 or fg_color == 0: # black fg and bright black fg fix
                        if bold:
                            fg_color = 9
                        else:
                            fg_color = 1
                    # 256 Colors
                if console:    
                    print(str(escape_codes), end="")

                # Add color to color map
                try:
                    new_frame.newColorMap[line_num][col_num] = [fg_color, bg_color]
                except Exception as E:
                    if console:    
                        print(str(E))
                        print(f"line num: {line_num}")
                i = end_index + 1
                continue    # jump the while
            elif text[end_index] == 'A':      # Move the cursor up X spaces
                escape_sequence = text[i + 2:end_index]
                if len(escape_sequence) == 0:
                    escape_sequence = 1
                move_by_amount = int(escape_sequence)
                line_num = line_num - move_by_amount
                i = end_index + 1
                continue    # jump the while
            elif text[end_index] == 'B':      # Move the cursor down X spaces
                escape_sequence = text[i + 2:end_index]
                if len(escape_sequence) == 0:
                    escape_sequence = 1
                move_by_amount = int(escape_sequence)
                line_num += move_by_amount
                i = end_index + 1
                continue    # jump the while
            elif text[end_index] == 'C':      # Move the cursor forward X spaces
                escape_sequence = text[i + 2:end_index]
                if len(escape_sequence) == 0:
                    escape_sequence = 1
                move_by_amount = int(escape_sequence)
                col_num += move_by_amount
                i = end_index + 1
                continue    # jump the while
            elif text[end_index] == 'D':      # Move the cursor back X spaces
                escape_sequence = text[i + 2:end_index]
                if len(escape_sequence) == 0:
                    escape_sequence = 1
                move_by_amount = int(escape_sequence)
                col_num = col_num - move_by_amount
                i = end_index + 1
                continue    # jump the while
            elif text[end_index] == 'H':      # Move the cursor to row/column
                escape_sequence = text[i + 2:end_index]
                escape_codes = escape_sequence.split(';')
                if len(escape_codes) > 1:   # row ; column
                    if escape_codes[0].isnumeric():
                        line_num = int(escape_codes[0])
                    if escape_codes[1].isnumeric():
                        col_num = int(escape_codes[1])
                elif len(escape_codes) == 1:   # row, column=1
                    #line_num = 1
                    if escape_codes[0].isnumeric():
                        col_num = int(escape_codes[0])
                i = end_index + 1
                continue    # jump the while
            elif text[end_index] == 'J':      # Clear screen
                # 0 or none = clear from cursor to end of screen
                # 1 = from cursor to top of screen
                # 2 = clear screen and move cursor to top left
                # 3 = clear entire screen and delete all lines saved in the scrollback buffer
                escape_sequence = text[i + 2:end_index]
                if len(escape_sequence) == 0:
                    escape_sequence = '0'   # default - clear from cursor to end
                if escape_sequence == '2':  # cls, move to top left
                    # using a sledgehammer to claer the screen
                    new_frame = durmovie.Frame(width, height + 1)
                    col_num, line_num = 0, 0
                i = end_index + 1   # move on to next byte
                continue
            elif text[end_index] == 's':    # save current position/state
                #saved_col_num = col_num
                #saved_line_num = line_num
                i = end_index + 1   # move on to next byte
                #saved_byte_location = i
                continue
            elif text[end_index] == 'u':    # restore saved position/state
                #col_num = saved_col_num
                #line_num = saved_line_num
                #i = saved_byte_location
                i = end_index + 1   # move on to next byte
                #pdb.set_trace()
                continue
            else:   # Some other escape code, who cares for now
                if appState.debug:
                    caller.notify(f"Unknown escape character type encountered: {text[end_index]}")
                i = end_index + 1   # move on to next byte
                continue
        # Or, not an escape character
        elif text[i] == '\n':   # new line (LF)
            line_num += 1
            if col_num > max_col:
                max_col = col_num
            col_num = 0
        elif text[i] == '\r':  # windows style newline (CR)
            pass    # pfft
        elif text[i] == '\x00': # Null byte
            pass
        elif text[i] == '\x1a': # ctl-z code, EOF, just before the SAUCE
            pass
        #elif text[i] == '\x01': # CTRL-A, SOH (start header).
        #    # Q: Why is this in some ANSIs? A: Because it's a smiley face in CP437
        #    pass
        #elif text[i] == '\x02': # CTRL-B, STX (start text)
        #    pass
        elif text[i:i + 5] == 'SAUCE' and len(text) - i == 128:   # SAUCE record found
            i += 128
        else:   # printable character (hopefully)
            if col_num >= maxWidth:
                col_num = 0
                line_num += 1
            character = text[i]
            try:
                new_frame.content[line_num][col_num] = character
            except IndexError:
                parse_error = True
                if debug:
                    caller.notify(f"Error writing content. Width: {width}, Height: {height}, line: {line_num}, col: {col_num}, char: {character}, pos: {i}")
            try:
                new_frame.newColorMap[line_num][col_num] = [fg_color, bg_color]
            except IndexError:
                parse_error = True
                if debug:
                    caller.notify(f"Error writing color. Width: {width}, Height: {height}, line: {line_num}, col: {col_num}, pos: {i}")
            if console:    
                print(character, end='')
            col_num += 1
        i += 1
    if console:    
        print("")
        print(f"Lines: {line_num}, Columns: {max_col}")
    if parse_error:
        caller.notify(f"Possible errors detected while loading this file. It may not display correctly.")
    height = line_num
    width = max_col
    frame = durmovie.Frame(height, width)
    line_num = 0
    col_num = 0
    # maybe usethis for the color: dur_ansilib.ansi_code_to_dur_16_color[fg_ansi]
    return new_frame


if __name__ == "__main__":
    # Example usage
    #file_path = 'kali.ans'
    #file_path = '11.ANS'
    file_path = '../rainbow.ans'
    if len(sys.argv) > 1:
        file_path = sys.argv[1]

    with open(file_path, 'r') as file:
        try:
            text_with_escape_codes = file.read()
        except UnicodeDecodeError:
            file.close()
            file = open(file_path, "r", encoding="cp437")
            #file = open(file_path, "r", encoding="big5")
            text_with_escape_codes = file.read()
        #parsed_text, fg, bg = parse_ansi_escape_codes(text_with_escape_codes)
        newFrame = parse_ansi_escape_codes(text_with_escape_codes,  console=True)
        print(str(newFrame.newColorMap))
        print(str(newFrame.content))
        print(str(newFrame))
        #print(parsed_text)
        #print(f"Fg: {fg}, bg: {bg}")





# Durdraw file operations - stuff related to open, save, etc

import datetime
import gzip
import os
import pdb
import pickle
import json

from durdraw.durdraw_options import Options
from durdraw.durdraw_movie import Movie
# import durdraw.durdraw_ansiparse as ansiparse

old_16_pal_to_new = {
    # maps old durdraw colors, found in file version 5 and lower
    # for 16 color
    0: 0,   # black
    1: 7,   # white/grey
    2: 3,   # cyan
    3: 5,   # magenta/purple
    4: 1,   # blue
    5: 6,   # brown/yellow
    6: 2,   # green
    7: 4,   # red
    8: 0,   # black
    9: 15,  # bright white
    10: 11, # bright cyan
    11: 13, # bright magenta
    12: 9,  # bright blue
    13: 14, # bright yellow
    14: 10, # bright green
    15: 12, # bright red
    16: 16, # bright black
}

old_256_pal_to_new = {
    # maps old durdraw colors, found in file version 5 and lower
    # for 16 color
    0: 0,   # black
    1: 7,   # white/grey
    2: 3,   # cyan
    3: 5,   # magenta/purple
    4: 1,   # blue
    5: 6,   # brown/yellow
    6: 2,   # green
    7: 4,   # red
    8: 241,   # bright black
    9: 8,  # bright red
    10: 9, # bright green
    11: 10, # bright yellow
    12: 11,  # bright blue
    13: 12, # bright magenta
    14: 13, # bright cyan
    15: 14, # bright white
    16: 15, # bright white
}

colors_to_html = {
    # Maps Durdraw color numbers to HTML values
    0: '#000000',   # black
    1: '#000000',   # black
    2: '#0000AA',   # blue
    3: '#00AA00',   # green
    4: '#00AAAA',   # cyan
    5: '#AA0000',   # red
    6: '#AA00AA',   # magenta
    7: '#AA5500',   # brown/dark yellow
    8: '#AAAAAA',   # light gray/white
    9: '#555555',   # dark grey
    10: '#5555FF',   # bright blue
    11: '#55FF55',   # bright green
    12: '#55FFFF',  # bright cyan
    13: '#FF5555',  # bright red
    14: '#FF55FF',  # bright magenta
    15: '#FFFF00',  # bright yellow
    16: '#FFFFFF',  # bright white
}

def write_frame_to_html_file(mov, appState, frame, file_path, gzipped=False):
    """ Writes a single frame to an HTML file
    """
    colorMode = appState.colorMode
    if gzipped:
        opener = gzip.open
    else:
        opener = open
    with opener(file_path, 'wt') as f:
        movieDataHeader = '<!DOCTYPE html>\n'
        movieDataHeader += '<html lang="en">\n'
        movieDataHeader += ''
        movieDataHeader += '    <head>\n'
        movieDataHeader += '        <meta charset="utf-8">\n'
        movieDataHeader += '        <meta name="viewport" content="width=device-width, initial-scale=1">\n'
        movieDataHeader += '        <title>DurDraw Generated Output</title>\n'
        movieDataHeader += '        <style>\n'
        movieDataHeader += '            body {\n'
        movieDataHeader += '                color: #FFFFFF;\n'
        movieDataHeader += '                background-color: #000000;\n'
        movieDataHeader += '            }\n'
        movieDataHeader += '            #duroutput {\n'
        movieDataHeader += '                font-size:12pt;\n'
        movieDataHeader += '                font-family: "Courier New", monospace;\n'
        # movieDataHeader += '              font-family: monospace,monospace;\n'
        movieDataHeader += '                line-height:13pt;\n'
        movieDataHeader += '                color: #FFFFFF;\n'
        movieDataHeader += '            }\n'
        movieDataHeader += '        </style>\n'
        movieDataHeader += '    </head>\n'
        movieDataHeader += '    <body>\n'
        movieDataHeader += '        <pre id="duroutput">\n'

        fileContent = movieDataHeader
        #lineNum = 0
        #colNum = 0
        newColorMap = []
        for posY in range(0, mov.sizeY):
            #newColorMap.append(list())
            for posX in range(0, mov.sizeX):
                #newColorMap[posX].append(list(frame.colorMap[posY, posX]))
                #newColorMap[posX].append(frame.newColorMap[posY][posX])
                durDolor = frame.newColorMap[posY][posX]
                fgColor = frame.newColorMap[posY][posX][0]
                bgColor = frame.newColorMap[posY][posX][1]
                c = frame.content[posY][posX]
        #for line in frame.content:
        #    for c in line: 
                try:
                    #fgColor = frame.newColorMap[colNum][lineNum][0]
                    #bgColor = frame.newColorMap[colNum][lineNum][1]
                    if appState.colorMode == "16":
                        bgColor += 1
                        if bgColor == 9:    # black duplicate
                            bgColor = 0
                    if appState.colorMode == "256":
                        fgColor += 1
                        bgColor += 1
                    if fgColor in colors_to_html.keys():
                        fgHtmlColor = colors_to_html[fgColor]
                    else:
                        fgHtmlColor = '#AAAAAA'
                    if bgColor in colors_to_html.keys():
                        bgHtmlColor = colors_to_html[bgColor]
                    else:
                        bgHtmlColor = '#000000'
                    #fileContent += f"<div style=\"color: {fgHtmlColor}; background-color: {bgHtmlColor}; display: inline;\">"
                    fileContent += f"<span style=\"color: {fgHtmlColor}; background-color: {bgHtmlColor};\">"
                    #fileContent += f"<font color={fgHtmlColor} style=\"background-color: {bgHtmlColor};\">"
                    #fileContent += f"<font color={fgHtmlColor} background-color={bgHtmlColor}>"
                    fileContent += str(c)
                    fileContent += "</span>"
                    #fileContent += "</div>"
                    #colNum += 1
                except Exception as e:
                    print(str(e))
                    pdb.set_trace()
            fileContent += '\n'
            #lineNum += 1
        
        fileContent += '        </pre>\n'
        fileContent += '    </body>\n'
        fileContent += '</html>\n'
        f.write(fileContent)
    f.close()

def serialize_to_json_file(opts, appState, movie, file_path, gzipped=True):
    """ Takes live Durdraw movie objects and serializes them out to a JSON files
    """
    colorMode = appState.colorMode
    if gzipped:
        opener = gzip.open
    else:
        opener = open
    with opener(file_path, 'wt') as f:
        movieDataHeader = {
            'formatVersion': opts.saveFileFormat,
            'colorFormat': colorMode, # 16, 256
            'preferredFont': 'fixed',   # fixed, vga, amiga, etc.
            'encoding': appState.charEncoding,
            'name': '',
            'artist': '',
            'framerate': opts.framerate,
            'sizeX': movie.sizeX,
            'sizeY': movie.sizeY,
            'extra': None,
            'frames': None,
            }
        frameNumber = 1
        fullMovie = {'DurMovie': movieDataHeader}
        fullMovieFrames = []
        for frame in movie.frames:
            content = ''
            newFrame = []
            newColorMap = []
            for posX in range(0, movie.sizeX):
                newColorMap.append(list())
                for posY in range(0, movie.sizeY):
                    #newColorMap[posX].append(list(frame.colorMap[posY, posX]))
                    try:
                        newColorMap[posX].append(frame.newColorMap[posY][posX])
                    except Exception as E:
                        print(E)
                        pdb.set_trace()
            for line in frame.content:
                content = ''.join(line)
                newFrame.append(content)
            serialized_frame = {
                'frameNumber': frameNumber,
                'delay': frame.delay,
                'contents': newFrame,
                'colorMap': newColorMap,
            }
            fullMovieFrames.append(serialized_frame)
            frameNumber += 1
        fullMovie['DurMovie']['frames'] = fullMovieFrames
        fullMovieJSON = json.dumps(fullMovie, indent=2)
        newMovieJSON = clean_up_json_output(fullMovieJSON)
        f.write(newMovieJSON)
    f.close()

def clean_up_json_output(json_str):
    """ Take aggressively whitespaced nested JSON lists and remove some of
    the excessive newlines and spaces. Basically format colorMap to
    look nicer
    """
    json_data = json_str
    json_data = json_data.replace("[\n            [", "[[")
    json_data = json_data.replace("[\n              ", "[")
    json_data = json_data.replace(",\n             ", ",")
    json_data = json_data.replace("\n            ],", "],")
    json_data = json_data.replace("],\n            [", "],[")
    json_data = json_data.replace("\n            ]", "]")
    json_data = json_data.replace("\n          ],", "],")
    return json_data

def get_file_mod_date_time(filename):
    """ Returns a string like this: 2009-10-06 10:50:01 """
    t = os.path.getmtime(filename)
    return str(datetime.datetime.fromtimestamp(t, tz=datetime.timezone.utc))[:19]


def get_dur_file_colorMode_and_charMode(f):
    """ Returns the color mode and encoding used by the file """
    f.seek(0)
    try:
        loadedMovieData = json.load(f)
    except Exception as e:
        return False
    colorMode = loadedMovieData['DurMovie']['colorFormat']
    try:
        charEncoding = loadedMovieData['DurMovie']['encoding']
    except Exception as e:
        charEncoding = 'utf-8'
    # convert from file format 5 to 6
    #if colorMode == 'xterm-256':
    #    colorMode = '256'
    if charEncoding == 'Utf-8':
        charEncoding = 'utf-8'
    return colorMode, charEncoding

def open_json_dur_file(f, appState):
    """ Loads json file into opts and movie objects.  Takes an open file
    object. Returns the opts and mov objects, encased in a list object.
     Or return False if the open fails.
    """
    f.seek(0)
    try:
        loadedMovieData = json.load(f)
    except Exception as e:
        return False

    width = loadedMovieData['DurMovie']['sizeX']
    height = loadedMovieData['DurMovie']['sizeY']
    colorMode = loadedMovieData['DurMovie']['colorFormat']
    newOpts = Options(width=width, height=height)
    newOpts.framerate = loadedMovieData['DurMovie']['framerate']
    newOpts.saveFileFormat = loadedMovieData['DurMovie']['formatVersion']
    # load frames into a new movie object
    newMov = Movie(newOpts)
    currentFrame = 0
    lineNum = 0
    for frame in loadedMovieData['DurMovie']['frames']:
        newMov.insertCloneFrame()
        newMov.nextFrame()
        for line in frame['contents']:
            #pdb.set_trace()
            if lineNum < len(newMov.frames[currentFrame].content):
                newMov.frames[currentFrame].content[lineNum] = [c for c in line]
                #newMov.frames[currentFrame].content[lineNum] = [c for c in line.encode(appState.charEncoding)]
                #newMov.frames[currentFrame].content[lineNum] = [c for c in bytes(line).decode('cp437')]
            #newMov.frames[currentFrame].content = line
            lineNum += 1
        lineNum = 0
        # load color map into new movie
        #pdb.set_trace()
        for x in range(0, width):
            for y in range(0, height):
                #pdb.set_trace()
                colorPair = frame['colorMap'][x][y]
                if appState.colorMode == '16' and colorMode == '256':
                    # set a default color when down-converting color modes:
                    if colorPair[0] > 16:
                        colorPair = [8,0]
                #if newOpts.saveFileFormat < 6:
                #    colorPair = convert_old_color_to_new(oldColorPair)
                #    if colorPair == None:
                #        pdb.set_trace()
                newMov.frames[currentFrame].colorMap[y, x] = tuple(colorPair)
                newMov.frames[currentFrame].newColorMap[y][x] = colorPair
        # Add delay for the frame
        newMov.frames[currentFrame].delay = frame['delay']
        currentFrame += 1
        #pdb.set_trace()
    newMov.deleteCurrentFrame()
    newMov.gotoFrame(1)
    container = {'opts':newOpts, 'mov':newMov}
    return container

def convert_old_color_to_new(oldPair, colorMode="16"):
    """ takes in a frame['colorMap'][x][y] list, returns a new list with
    [0] replaced by appropriately mapped # """
    #newPair = oldPair
    newFg = oldPair[0]
    newBg = oldPair[1]
    if colorMode == "16":
        if oldPair[0] in old_16_pal_to_new.keys():
            newFg = old_16_pal_to_new[oldPair[0]]
            newFg += 1
        if oldPair[1] in old_16_pal_to_new.keys():
            newBg = old_16_pal_to_new[oldPair[1]]
            #newBg += 1
    if colorMode == "256":
        if oldPair[0] in old_256_pal_to_new.keys():
            newFg = old_256_pal_to_new[oldPair[0]]
            newFg += 1
        #if oldPair[1] in old_16_pal_to_new.keys():
        #    newBg = old_16_pal_to_new[oldPair[1]]
        newBg = oldPair[1]    # 256 does not use bg color, so don't change it
    #pdb.set_trace()
    return [newFg, newBg]


class DurUnpickler(pickle.Unpickler):
    """" Custom Unpickler to remove serialized module names (like __main__) from
    unpickled data and replace them with "durdraw.durdraw_movie"
    """
    def find_class(self, module, name):
        #if module == "__main__":
        module = "durdraw.durdraw_movie"
        return super().find_class(module, name)


 
